# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/ml.vision.ipynb.

# %% auto 0
__all__ = ['Make3Channel', 'TorchVisionTransform']

# %% ../../nbs/ml.vision.ipynb 3
from PIL.Image import Image 
from torchvision.transforms import Compose

from fastcore.transform import Transform
from fastai.vision.core import PILImage

# %% ../../nbs/ml.vision.ipynb 4
class Make3Channel:
    """Tiles 1 channel image to 3 channel"""

    def __call__(self, x):
        if isinstance(x, Image):
            return x.convert(mode='RGB')
        rpts = (3, 1, 1) if x.ndim == 3 else (1, 3, 1, 1)
        return x.repeat(*rpts)
    
    def __repr__(self):
        return "Make3Channel()"

# %% ../../nbs/ml.vision.ipynb 5
class TorchVisionTransform(Transform):
    """Converts a torchvision transform to fastai transform"""
    def __init__(self, transform):
        self.tfm = transform

    def encodes(self, o: PILImage): 
        return self.tfm(o)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/jerx.reward.gpt.ipynb.

# %% auto 0
__all__ = ['log', 'DEFAULT_SYSTEM_PROMPT', 'USER_PROMPT', 'DEFAULT_PROMPT_MESSAGES', 'Result', 'make_question_answer_func',
           'make_reward_func']

# %% ../../../nbs/jerx.reward.gpt.ipynb 3
from langchain.chat_models import ChatOpenAI
from langchain.prompts import ChatPromptTemplate
from langchain.chains.openai_functions import create_structured_output_runnable
from langchain_core.pydantic_v1 import BaseModel, Field

from ...text.utils import fuzzy_match
from ...logging import get_logger

log = get_logger(__name__)

# %% ../../../nbs/jerx.reward.gpt.ipynb 5
DEFAULT_SYSTEM_PROMPT = """You are an expert Q&A system that is trusted around the world. You are given a question that requires multi-hop reasoning. Always answer the question using the provided context information, and not prior knowledge.
Some rules to follow:
1. Never directly reference the given context in your answer.
2. Avoid statements like 'Based on the context, ...' or 'The context information ...' or anything along those lines.
"""

USER_PROMPT = """The context information below is provided as a set of entity-relation-entity triplets from knowledge graph.
---------------------
{context}
---------------------
Given the context information and not prior knowledge, answer the question.
{question}
"""

DEFAULT_PROMPT_MESSAGES = ChatPromptTemplate.from_messages([
    ("system", DEFAULT_SYSTEM_PROMPT), 
    ("user", USER_PROMPT),
])


class Result(BaseModel):
    """Data model for answering the question."""
    answer: str = Field(description="The answer to the question in 2-4 words.")
    reasoning: str = Field(description="Multi-hop reasoning for the answer.")


def make_question_answer_func(llm: ChatOpenAI | None = None):
    if llm is None:
        llm = ChatOpenAI(model="gpt-3.5-turbo-1106", temperature=0)

    chain = create_structured_output_runnable(Result, llm=llm, prompt=DEFAULT_PROMPT_MESSAGES)
    def func(context: str, question: str) -> Result:
        return chain.invoke(dict(context=context, question=question))
    return func

# %% ../../../nbs/jerx.reward.gpt.ipynb 6
def make_reward_func(llm: ChatOpenAI | None = None, answer_comparator=fuzzy_match):
    qa = make_question_answer_func(llm)

    def reward(context: str, question: str, answers: list[str]) -> float:
        result = qa(context, question)
        correct = any(answer_comparator(result.answer, answer) for answer in answers)
        return int(correct)

    return reward



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/ml.loss.ipynb.

# %% auto 0
__all__ = ['BinaryFocalLoss']

# %% ../../nbs/ml.loss.ipynb 3
import torch
from torch import nn
from fastcore.basics import ifnone

# %% ../../nbs/ml.loss.ipynb 4
# license: https://www.kaggle.com/crissallan/pytorchlightning-efficientnet-focalloss-training
class BinaryFocalLoss(nn.Module):
    """
    <sup>
    T.-Y. Lin, P. Goyal, R. Girshick, K. He, and P. Dollar, “Focal loss for dense object detection,” in Proceedings of the IEEE International Conference on Computer Vision (ICCV), Oct 2017.
    </sup>
    """
    def __init__(self, alpha=1, gamma=2, weight=None, reduction=torch.mean, thresh=0.5):
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.weight = None if weight is None else torch.tensor([weight], dtype=torch.float, requires_grad=False, device='cuda')
        self.reduction = reduction
        self.thresh = 0.5
        self.eps = 1e-12 # prevent training from Nan-loss error

    def forward(self, logits, targets):
        """
        logits & target should be tensors with shape [batch_size, num_classes]
        """
        probs = torch.sigmoid(logits)
        log_pt = targets * torch.log(probs + self.eps) + (1.0 - targets) * torch.log(1.0 - probs + self.eps)
        focal_loss = -self.alpha * ((1 - torch.exp(log_pt)) ** self.gamma) * log_pt
        if self.weight is not None:
            focal_loss = focal_loss * self.weight
        
        reduction = ifnone(self.reduction, lambda x:x)
        return reduction(focal_loss)
    
    def decodes(self, x):    
        return x > self.thresh

    def activation(self, x): 
        return torch.sigmoid(x)
